---
layout: post
title: 如何用PHP制作一个简单的反向代理
tags: [PHP, 代理]
---

  解决问题还是很简单的。<!--more-->    

# 起因
  由于莫名其妙的原因，很多运营商都屏蔽了Github Pages，导致我上我的博客很困难。这令我很不高兴，但是没办法，毕竟在中国就是这样，搭个网站很麻烦。   
  在两年前，我的博客也被屏蔽过一次，那时候我是用的Nginx反代解决这个问题的，不过Nginx反代要服务器，当时我用的是花火学园的服务器来搞，但是现在因为登一次花火学园的服务器很麻烦，网络卡的不得了，所以没办法，我只好想想别的更简单的方法解决这个问题。   
  这时候我就想到了PHP，用PHP解决这个问题应该很简单，用`file_get_contents()`应该很轻松就可以解决这个问题。   
  
# 解决过程
  理论上来说，只要整一个php然后里面直接用`echo file_get_contents(url)`就可以实现反代了，所以首先就按这个思路做，为了让访问的所有请求都发送到这个php里，我得用重写的方式。   
  我用的是GearHost的虚拟主机，他们的主机都用的是IIS，这导致我还得了解一下IIS的重写怎么搞，还好我的方案和WordPress的重写方案很像，我就直接按照WordPress的方式写了。   
  把下面的代码放到Rewrite的rules段里就可以正常工作了：
```xml
<rule name="Mayx" stopProcessing="true">
<match url="^(.*)$" />
<conditions>
<add input="{REQUEST_FILENAME}" matchType="IsFile" negate="true" />
<add input="{REQUEST_FILENAME}" matchType="IsDirectory" negate="true" />
</conditions>
<action type="Rewrite" url="index.php"/>
</rule>
```
  然后就是新建一个index.php文件，路径的话用`$_SERVER['REQUEST_URI']`就可以了。   
  不过如果直接用`echo file_get_contents(url)`的方法搞会出现一些问题，比如css和js的MIME类型要求必须和实际一样，但是我这个方案会导致所有请求都是`text/html`类型的，这是个很麻烦的问题啊……   
  后来我花了2个小时查资料，搜php怎么读MIME，结果搜到的全是`mime_content_type()`之类乱七八糟的东西，让人很难受。我仔细思考了一下，MIME是在Header里声明的，我应该去搜Header，而不是怎么读MIME，最终可算是搜到了一点有用的东西了，也就是`$http_response_header`，这个东西好像是个数组，后来我想了一下干脆不要考虑怎么读MIME，直接把Header跟着重写一遍算了，于是最终代码如下：   
```php
<?php
$content=file_get_contents("https://mabbs.github.io".$_SERVER['REQUEST_URI']);
foreach($http_response_header as $header){
header($header);
}
if($content){
echo $content;
}else{
echo "404";
}
```

# 之后的操作
  这么搞完之后基本上就可以正常工作了，于是我的博客域名再一次更换（专供中国用户）：[yuki.gear.host](https://yuki.gear.host/)，另外把备用的Gitalk的Github Apps返回地址也改了，基本上就可以正常工作了。希望以后GFW能做个人，不要再乱搞国内的网络环境了。
